import math
from tqdm import tqdm
from langchain_community.embeddings import HuggingFaceBgeEmbeddings


class BGEEmbedModel:

    def __init__(self, batch_size=500, device="cpu", model_name="BAAI/bge-small-en-v1.5"):
        self.batch_size = batch_size
        self.model = HuggingFaceBgeEmbeddings(
            model_name=model_name, model_kwargs={"device": device}, encode_kwargs={"normalize_embeddings": True}
        )

    def embed_documents(self, documents):
        embeddings = []
        num_batch = math.ceil(len(documents) / self.batch_size)
        for idx in tqdm(range(0, len(documents), self.batch_size), total=num_batch):
            batch_documents = documents[idx: idx+self.batch_size]
            batch_embeddings = self.model.embed_documents(batch_documents)
            embeddings.extend(batch_embeddings)
        return embeddings
    
    def embed_query(self, document):
        return self.model.embed_query(document)


import time

hdbscan_params = {
    # preprocessing part
    'preprocessing': {'words_range': (1, 500)},
    # name filter part
    'name_filter': {},
    # extracting keywords part
    'extract_keywords': {'llm_model': 'gpt-35-turbo', 'temperature': 0., 'batch_size': 300},
    # embedding part
    'embedding': {'model': 'bge', 'batch_size': 500, 'device': 'mps'},
    # hdbscan clustering part
    'hdbscan': {'reduced_dim': 5, 'n_neighbors': 10, 'min_cluster_percent': 0.02, 'topk': 5, 
                'llm_model': 'gpt-35-turbo', 'temperature': 0.5, 'batch_size': 300, 'verbose': True},
}

kmeans_params = {
    # preprocessing part
    'preprocessing': {'words_range': (1, 500)},
    # name filter part
    'name_filter': {},
    # embedding part
    'embedding': {'model': 'bge', 'batch_size': 500, 'device': 'mps'},
    # hdbscan clustering part
    'kmeans': {'n_clusters_list': [100, 30, 5], 'topk': 5, 'llm_model': 'gpt-35-turbo', 'temperature': 0.5,
               'batch_size': 300, 'embed_model': "bge", "device": "mps", "embed_batch_size": 500,
               'ngram_range': (1, 2), 'topk_keywords': 10},
}


def topic_modeling_by_hdbscan(data, text_col_name, params=hdbscan_params):
    tt1 = time.time()
    if text_col_name not in data.columns:
        raise ValueError("`text_col_name` should be in data.columns")
    data_df = data.rename(columns={text_col_name: 'input'})
    data_df = data_df.dropna(subset=["input"])
    data_df['input'] = data_df['input'].astype(str)
    print(f"#Original data size: {len(data_df)}")

    # preprocessing
    if 'preprocessing' in params:
        print("-"*30, "Preprocessing part", "-"*30)
        t1 = time.time()
        from topicgpt.preprocessing import MinMaxLengthFilter
        filter = MinMaxLengthFilter(words_range=params['preprocessing']['words_range'])
        data_df = filter.transform(data_df)
        t2 = time.time()
        print(f"#Data size after preprocessing: {len(data_df)}, processing_time: {t2-t1:0.1f} s")


    if 'name_filter' in params:
        print("-"*30, "Name_filter part", "-"*30)
        t1 = time.time()
        from topicgpt.preprocessing import NameFilter
        filter = NameFilter()
        data_df = filter.transform(data_df)
        t2 = time.time()
        print(f"#Data size after name_filter: {len(data_df)}, preprocessing_time: {t2-t1:0.1f} s")

    if 'extract_keywords' in params:
        print("-"*30, "Extracting keywords part", "-"*30)
        t1 = time.time()
        from topicgpt.feature_extraction import LlmKeywordsExtractor
        llm = LlmKeywordsExtractor(model_name=params['extract_keywords']['llm_model'], 
                                   temperature=params['extract_keywords']['temperature'], 
                                   batch_size=params['extract_keywords']['batch_size'])
        data_df = llm.transform(data=data_df)
        t2 = time.time()
        print(f"#Data size after extracting_filter: {len(data_df)}, preprocessing_time: {t2-t1:0.1f} s")
        
    if 'embedding' in params:
        print("-"*30, "Embeddding part", "-"*30)
        t1 = time.time()
        if params['embedding']['model'] == 'ada':
            from topicgpt.walmart_llm import AdaEmbedModel
            llm = AdaEmbedModel(batch_size=params['embedding']['batch_size'])
        elif params['embedding']['model'] == 'bge':
            from topicgpt.embedding import BGEEmbedModel
            llm = BGEEmbedModel(batch_size=params['embedding']['batch_size'], device=params['embedding']['device'])
            
        embeddings = llm.embed_documents(data_df['input'].tolist())
        data_df['embeddings'] = embeddings
        t2 = time.time()
        print(f"#Data size after embedding: {len(data_df)}, preprocessing_time: {t2-t1:0.1f} s")
        
    else:
        raise ValueError("`embedding` is a neccessary part of this pipeline.")

    if 'hdbscan' in params:
        print("-"*30, "HDBSCAN clustering part", "-"*30)
        t1 = time.time()
        from topicgpt.topic import HDBSCANTopicGenerator
        model = HDBSCANTopicGenerator(reduced_dim=params['hdbscan']['reduced_dim'], 
                                      n_neighbors=params['hdbscan']['n_neighbors'], 
                                      min_cluster_percent=params['hdbscan']['min_cluster_percent'], 
                                      topk=params['hdbscan']['topk'],
                                      model_name=params['hdbscan']['llm_model'], 
                                      temperature=params['hdbscan']['temperature'], 
                                      verbose=params['hdbscan']['verbose'])
        tree_root = model.predict(data_df)
        t2 = time.time()
        print(f"#Data size after hdbscan: {len(data_df)}, preprocessing_time: {t2-t1:0.1f} s")
    else:
        raise ValueError("`hdbscan` is a neccessary part of this pipeline.")

    tt2 = time.time()
    print(f"Total processing time: {tt2-tt1:0.1f} s")
    return tree_root


def topic_modeling_by_kmeans(data, text_col_name, params=kmeans_params):
    tt1 = time.time()
    if text_col_name not in data.columns:
        raise ValueError("`text_col_name` should be in data.columns")
    data_df = data.rename(columns={text_col_name: 'input'})
    data_df = data_df.dropna(subset=["input"])
    data_df['input'] = data_df['input'].astype(str)
    print(f"#Original data size: {len(data_df)}")

    # preprocessing
    if 'preprocessing' in params:
        print("-"*30, "Preprocessing part", "-"*30)
        t1 = time.time()
        from topicgpt.preprocessing import MinMaxLengthFilter
        filter = MinMaxLengthFilter(words_range=params['preprocessing']['words_range'])
        data_df = filter.transform(data_df)
        t2 = time.time()
        print(f"#Data size after preprocessing: {len(data_df)}, processing_time: {t2-t1:0.1f} s")


    if 'name_filter' in params:
        print("-"*30, "Name_filter part", "-"*30)
        t1 = time.time()
        from topicgpt.preprocessing import NameFilter
        filter = NameFilter()
        data_df = filter.transform(data_df)
        t2 = time.time()
        print(f"#Data size after name_filter: {len(data_df)}, preprocessing_time: {t2-t1:0.1f} s")

    if 'embedding' in params:
        print("-"*30, "Embeddding part", "-"*30)
        t1 = time.time()
        if params['embedding']['model'] == 'ada':
            from topicgpt.walmart_llm import AdaEmbedModel
            llm = AdaEmbedModel(batch_size=params['embedding']['batch_size'])
        elif params['embedding']['model'] == 'bge':
            from topicgpt.embedding import BGEEmbedModel
            llm = BGEEmbedModel(batch_size=params['embedding']['batch_size'], device=params['embedding']['device'])
            
        embeddings = llm.embed_documents(data_df['input'].tolist())
        data_df['embeddings'] = embeddings
        t2 = time.time()
        print(f"#Data size after embedding: {len(data_df)}, preprocessing_time: {t2-t1:0.1f} s")
        
    else:
        raise ValueError("`embedding` is a neccessary part of this pipeline.")

    if 'kmeans' in params:
        print("-"*30, "Kmeans Clustering part", "-"*30)
        t1 = time.time()
        from topicgpt.topic import KmeansTopicGenerator
        model = KmeansTopicGenerator(n_clusters_list=params['kmeans']['n_clusters_list'], 
                                     topk=params['kmeans']['topk'], 
                                     model_name=params['kmeans']['llm_model'],
                                     batch_size=params['kmeans']['batch_size'],
                                     temperature=params['kmeans']['temperature'], 
                                     embed_model=params['kmeans']['embed_model'],
                                     device=params['kmeans']['device'], 
                                     embed_batch_size=params['kmeans']['embed_batch_size'])
        clusters = []
        microclusters = model.generate_micro_clusters(data_df)

        from topicgpt.feature_extraction import TfidifKeywordsExtractor

        extractor = TfidifKeywordsExtractor()
        clusters.append(extractor.transform(microclusters))

        for idx, _ in enumerate(params['kmeans']['n_clusters_list'][1:], start=1):
            clusters.append(model.generate_higher_clusters(clusters[-1], level=idx))
        t2 = time.time()
        print(f"#Data size after hdbscan: {len(data_df)}, preprocessing_time: {t2-t1:0.1f} s")
    else:
        raise ValueError("`hdbscan` is a neccessary part of this pipeline.")

    tt2 = time.time()
    print(f"Total processing time: {tt2-tt1:0.1f} s")
    return clusters


import time
import random
import tiktoken
from ..base import TransformerMixin
from ..walmart_llm import ChatModel

class LlmSummaryExtractor(TransformerMixin):

    def __init__(self, output_len, model_name="gpt-35-turbo", temperature=0., batch_size=500):
        self.output_len = output_len
        self.model_name = model_name
        self.template = "Summarize following sentences within {output_len} words:\n{text}"
        self.model = ChatModel(model_name=model_name, temperature=temperature, batch_size=batch_size)
        self.encoding = tiktoken.encoding_for_model(model_name=model_name)
    
    def transform(self, texts, topk=5, iterations=100, **params):
        prompts = []
        for iter in range(iterations):
            input_text = ""
            for idx in range(topk):
                if (self.model_name == "gpt-35-turbo" and len(self.encoding.encode(input_text)) > 4000) or \
                   (self.model_name == "gpt-4" and len(self.encoding.encode(input_text)) > 7000):
                    break
                input_text += f"sentence {idx}: {random.choice(texts).strip()}\n"
            prompts.append(self.template.format(output_len=self.output_len, text=input_text))
        
        s1 = time.time()
        responses = []
        for prompt in prompts:
            try:
                responses.append(self.model.completion(prompt))
            except:
                continue
        s2 = time.time()
        print(f"Time: {s2 - s1:.02f}s")
        return responses


import abc


class TransformerMixin(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def transform(self, texts, **params):
        pass


class GeneratorMixin(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def predict(self, texts, embeddings, **params):
        pass


import tiktoken
from ..base import TransformerMixin
from ..walmart_llm import ChatModel
from ..topic import Cluster


class TopicGenerator(TransformerMixin):

    def __init__(self, model_name="gpt-35-turbo", temperature=0.5, batch_size=100):
        self.model_name = model_name
        self.model = ChatModel(model_name=model_name, temperature=temperature, batch_size=batch_size)
        self.encoding = tiktoken.encoding_for_model(model_name=model_name)

    def _generate_topic_and_description_for_cluster(self, cluster):
        prefix = """You are a good editor. Given some requirements and sentences, you need to summarize the main topic of those sentences and give a simple description of this topic according to those requirements. 
        Here are some requirements you MUST follow:
        1. The topic should be less than 10 words.
        2. The description should be less than 30 words.

        Here are the sentences you need to consider all:
        {user_input}
        """
        
        suffix = """The output should be in the json format:
        {"topic": <summarize the main topic of those sentences>, "description": <output a discription for this topic>}
        """
        user_input = ""
        for i, text in enumerate([cluster.data['input'][idx] for idx in cluster.closest], start=1):
            if (self.model_name == "gpt-35-turbo" and len(self.encoding.encode(user_input)) > 3000) or \
               (self.model_name == "gpt-4" and len(self.encoding.encode(user_input)) > 7000):
                break
            user_input += f"sentence {i}: {str(text).strip()}\n"
        prompt = prefix.format(user_input=user_input) + suffix
        try:
            return self.model.retry_json_completeion(prompt, keys=['topic', 'description'])
        except Exception as e:
            print(f"Exception: {e}")
            return "{'topic': 'OpenAI policy violation', 'description': ''}"

    def _generate_batch_topics_and_descriptions_for_clusters(self, clusters):
        prefix = """Provide a concise topic and a corresponding description of the triple backquoted sentences.
```{user_input}```
"""
        suffix = """
The output should be in the json format:
{"topic": <summarize the topic of those sentences>, "description": <output a discription for this topic at most 50 words>}"""

        prompts = []
        for cluster in clusters:
            user_input = ""
            for i, text in enumerate([cluster.data['input'][idx] for idx in cluster.closest], start=1):
                if (self.model_name == "gpt-35-turbo" and len(self.encoding.encode(user_input)) > 3000) or \
                   (self.model_name == "gpt-4" and len(self.encoding.encode(user_input)) > 7000):
                    break
                user_input += f"sentence {i}: {str(text).strip()}\n"
                # user_input += f"""sentence {i}: ""{str(text).strip()}"""""
            prompts.append(prefix.format(user_input=user_input)+suffix)
        
        responses = self.model.batch_completion(prompts) 
        for idx, (prompt, response) in enumerate(zip(prompts, responses)):
            try:
                topic = eval(response)['topic']
                description = eval(response)['description']
            except Exception as e:
                try:
                    responses[idx] = self.model.retry_json_completeion(prompt, keys=['topic', 'description'])
                except:
                    responses[idx] = "{'topic': 'OpenAI policy violation or Not JSON format', 'description': 'violation'}"
        return responses

    def _generate_batch_topics_and_descriptions_for_higher_clusters(self, clusters):
        prefix = """You're a good editor. Given several subtopics, descriptions and keywords, your task is to summarize a topic that try to encompass most subtopics and provide a description of that topic according to following requirments.

        Here are some requirements that you must adhere to:
        1. The topic MUST be less than 10 words.
        2. The description MUST be less than 50 words.
        3. DON'T output person's name.

        Here are the subtopics, descriptions and keywords:
        """
        suffix = """Output the topic and description in the json format as below:
        {"topic": [output summarized topic here], "description": [output detailed description here]}"""

        prompts = []
        for cluster in clusters:
            lower_topics_str = ""
            for child in cluster.children:
                lower_topics_str += f"Subtopic: {child.topic}\n"
                lower_topics_str += f"Description: {child.description}\n"    
            prompts.append(prefix + lower_topics_str + suffix)
        
        responses = self.model.batch_completion(prompts) 
        for idx, (prompt, response) in enumerate(zip(prompts, responses)):
            try:
                topic = eval(response)['topic']
                description = eval(response)['description']
            except Exception as e:
                try:
                    responses[idx] = self.model.retry_json_completeion(prompt, keys=['topic', 'description'])
                except:
                    responses[idx] = "{'topic': 'OpenAI policy violation or Not JSON format', 'description': 'violation'}"
        return responses

    def transform(self, obj):
        if isinstance(obj, list):
            return self._generate_batch_topics_and_descriptions_for_clusters(obj)
        elif isinstance(obj, Cluster):
            return self._generate_topic_and_description_for_cluster(obj)
    



from setuptools import setup, find_packages

with open("requirements.txt", "r") as f:
    install_requires = f.read().splitlines()

with open("topicgpt/README.md", "r") as f:
    long_description = f.read()

setup(
    name='wm_topicgpt',
    version='0.1.0',
    description='This is a package to generate topics for the text corpus.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    install_requires=install_requires,
    python_requires='>=3.9',
)

# upload codes
# python3 setup.py sdist bdist_wheel
# twine upload dist/*


import pandas as pd
from typing import List, Any
from dataclasses import dataclass, field


@dataclass
class Cluster:
    data: pd.DataFrame = None
    children: List[Any] = field(default_factory=list)

    topic: str = None
    description: str = None
    keywords: List[str] = None

    size: int = None
    percent: float = None

    centroid: List[float] = None
    closest: List[str] = None
    within_cluster_dist: int = None